@@ -1,16 +1,18 @@
-package HTTP::Server::Simple;
-use 5.006;
 use strict;
 use warnings;
+
+package HTTP::Server::Simple;
+use FileHandle;
 use Socket;
 use Carp;
+use URI::Escape;
 
-our $VERSION = '0.20';
+use vars qw($VERSION $bad_request_doc);
+$VERSION = '0.38';
 
 =head1 NAME
 
-HTTP::Server::Simple
-
+HTTP::Server::Simple - Lightweight HTTP server
 
 =head1 SYNOPSIS
 
@@ -40,14 +42,13 @@ module (see L<HTTP::Server::Simple::CGI>);
 
 =head1 DESCRIPTION
 
-This is a simple standalone HTTP server. By default, it doesn't thread 
-or fork.
-
-It does, however, act as a simple frontend which can be used 
+This is a simple standalone HTTP server. By default, it doesn't thread
+or fork. It does, however, act as a simple frontend which can be used
 to build a standalone web-based application or turn a CGI into one.
 
-(It's possible to use Net::Server to get threading, forking,
-preforking and so on. Autrijus Tang wrote the functionality and owes docs for that ;)
+It is possible to use L<Net::Server> classes to create forking,
+pre-forking, and other types of more complicated servers; see
+L</net_server>.
 
 By default, the server traps a few signals:
 
@@ -55,26 +56,79 @@ By default, the server traps a few signals:
 
 =item HUP
 
-When you C<kill -HUP> the server, it does its best to rexec
-itself.  Please note that in order to provide restart-on-SIGHUP,
-HTTP::Server::Simple sets a SIGHUP handler during initialisation. If
-your request handling code forks you need to make sure you reset this
-or unexpected things will happen if somebody sends a HUP to all running
-processes spawned by your app (e.g. by "kill -HUP <script>")
-
+When you C<kill -HUP> the server, it lets the current request finish being
+processed, then uses the C<restart> method to re-exec itself. Please note that
+in order to provide restart-on-SIGHUP, HTTP::Server::Simple sets a SIGHUP
+handler during initialisation. If your request handling code forks you need to
+make sure you reset this or unexpected things will happen if somebody sends a
+HUP to all running processes spawned by your app (e.g. by "kill -HUP <script>")
 
 =item PIPE
 
 If the server detects a broken pipe while writing output to the client, 
 it ignores the signal. Otherwise, a client closing the connection early 
-could kill the server
+could kill the server.
 
 =back
 
+=head1 EXAMPLE
+ 
+ #!/usr/bin/perl
+ {
+ package MyWebServer;
+ 
+ use HTTP::Server::Simple::CGI;
+ use base qw(HTTP::Server::Simple::CGI);
+ 
+ my %dispatch = (
+     '/hello' => \&resp_hello,
+     # ...
+ );
+ 
+ sub handle_request {
+     my $self = shift;
+     my $cgi  = shift;
+   
+     my $path = $cgi->path_info();
+     my $handler = $dispatch{$path};
+ 
+     if (ref($handler) eq "CODE") {
+         print "HTTP/1.0 200 OK\r\n";
+         $handler->($cgi);
+         
+     } else {
+         print "HTTP/1.0 404 Not found\r\n";
+         print $cgi->header,
+               $cgi->start_html('Not found'),
+               $cgi->h1('Not found'),
+               $cgi->end_html;
+     }
+ }
+ 
+ sub resp_hello {
+     my $cgi  = shift;   # CGI.pm object
+     return if !ref $cgi;
+     
+     my $who = $cgi->param('name');
+     
+     print $cgi->header,
+           $cgi->start_html("Hello"),
+           $cgi->h1("Hello $who!"),
+           $cgi->end_html;
+ }
+ 
+ } 
+ 
+ # start the server on port 8080
+ my $pid = MyWebServer->new(8080)->background();
+ print "Use 'kill $pid' to stop server.\n";
+
+=head1 METHODS
+
 =head2 HTTP::Server::Simple->new($port)
 
 API call to start a new server.  Does not actually start listening
-until you call C<-E<gt>run()>.
+until you call C<-E<gt>run()>.  If omitted, C<$port> defaults to 8080.
 
 =cut
 
@@ -91,19 +145,14 @@ sub new {
     bless( $self, $class );
     $self->port( $port || '8080' );
 
-
-
     return $self;
 }
 
 
 =head2 lookup_localhost
 
-Looks up the local host's hostname and IP address.
-
-Stuffs them into
-
-$self->{'localname'} and $self->{'localaddr'}
+Looks up the local host's IP address, and returns it.  For most hosts,
+this is C<127.0.0.1>.
 
 =cut
 
@@ -117,8 +166,6 @@ sub lookup_localhost {
 }
 
 
-
-
 =head2 port [NUMBER]
 
 Takes an optional port number for this server to listen on.
@@ -150,63 +197,46 @@ sub host {
 
 }
 
-=head2 background
+=head2 background [ARGUMENTS]
 
-Run the server in the background. returns pid.
+Runs the server in the background, and returns the process ID of the
+started process.  Any arguments will be passed through to L</run>.
 
 =cut
 
 sub background {
     my $self  = shift;
     my $child = fork;
-    die "Can't fork: $!" unless defined($child);
+    croak "Can't fork: $!" unless defined($child);
     return $child if $child;
 
     if ( $^O !~ /MSWin32/ ) {
         require POSIX;
         POSIX::setsid()
-            or die "Can't start a new session: $!";
+            or croak "Can't start a new session: $!";
     }
-    $self->run();
+    $self->run(@_);
 }
 
-=head2 run
+=head2 run [ARGUMENTS]
 
 Run the server.  If all goes well, this won't ever return, but it will
-start listening for http requests.
+start listening for C<HTTP> requests.  Any arguments passed to this
+will be passed on to the underlying L<Net::Server> implementation, if
+one is used (see L</net_server>).
 
 =cut
 
 my $server_class_id = 0;
 
+use vars '$SERVER_SHOULD_RUN';
+$SERVER_SHOULD_RUN = 1;
+
 sub run {
     my $self   = shift;
     my $server = $self->net_server;
 
-    # Handle SIGHUP
-
     local $SIG{CHLD} = 'IGNORE';    # reap child processes
-    local $SIG{HUP} = sub {
-
-        # XXX TODO: Autrijus says this code was incorrect when he wrote
-        # it and we should move to the sample code from perldoc perlipc
-        close HTTPDaemon;
-
-        # and then, on systems implementing fork(), we make sure
-        # we are running with a new pid, so another -HUP will still
-        # work on the new process.
-        require Config;
-        if ( $Config::Config{d_fork} and my $pid = fork() ) {
-
-            # finally, allow ^C on the parent process to terminate
-            # the children.
-            waitpid( $pid, 0 );
-            exit;
-        }
-
-        # do the exec. if $0 is not executable, try running it with $^X.
-        exec {$0}( ( ( -x $0 ) ? () : ($^X) ), $0, @ARGV );
-    } if exists $SIG{'HUP'};
 
     # $pkg is generated anew for each invocation to "run"
     # Just so we can use different net_server() implementations
@@ -219,7 +249,14 @@ sub run {
     if ($server) {
         require join( '/', split /::/, $server ) . '.pm';
         *{"$pkg\::ISA"} = [$server];
-        $self->print_banner;
+
+        # clear the environment before every request
+        require HTTP::Server::Simple::CGI;
+        *{"$pkg\::post_accept"} = sub {
+            HTTP::Server::Simple::CGI::Environment->setup_environment;
+            # $self->SUPER::post_accept uses the wrong super package
+            $server->can('post_accept')->(@_);
+        };
     }
     else {
         $self->setup_listener;
@@ -227,13 +264,15 @@ sub run {
         *{"$pkg\::run"} = $self->_default_run;
     }
 
-    $pkg->run( port => $self->port );
+    local $SIG{HUP} = sub { $SERVER_SHOULD_RUN = 0; };
+
+    $pkg->run( port => $self->port, @_ );
 }
 
 =head2 net_server
 
-User-overridable method. If you set it to a C<Net::Server> subclass,
-that subclass is used for the C<run> method.  Otherwise, a minimal 
+User-overridable method. If you set it to a L<Net::Server> subclass,
+that subclass is used for the C<run> method.  Otherwise, a minimal
 implementation is used as default.
 
 =cut
@@ -249,11 +288,11 @@ sub _default_run {
 
         $self->print_banner;
 
-        while (1) {
+        while ($SERVER_SHOULD_RUN) {
             local $SIG{PIPE} = 'IGNORE';    # If we don't ignore SIGPIPE, a
                  # client closing the connection before we
                  # finish sending will cause the server to exit
-            while ( accept( my $remote, HTTPDaemon ) ) {
+            while ( accept( my $remote = new FileHandle, HTTPDaemon ) ) {
                 $self->stdio_handle($remote);
                 $self->lookup_localhost() unless ($self->host);
                 $self->accept_hook if $self->can("accept_hook");
@@ -267,9 +306,38 @@ sub _default_run {
                 close $remote;
             }
         }
+
+        # Got here? Time to restart, due to SIGHUP
+        $self->restart;
     };
 }
 
+=head2 restart
+
+Restarts the server. Usually called by a HUP signal, not directly.
+
+=cut
+
+sub restart {
+    my $self = shift;
+
+    close HTTPDaemon;
+
+    $SIG{CHLD} = 'DEFAULT';
+    wait;
+
+    ### if the standalone server was invoked with perl -I .. we will loose
+    ### those include dirs upon re-exec. So add them to PERL5LIB, so they
+    ### are available again for the exec'ed process --kane
+    use Config;
+    $ENV{PERL5LIB} .= join $Config{path_sep}, @INC;
+
+    # Server simple
+    # do the exec. if $0 is not executable, try running it with $^X.
+    exec {$0}( ( ( -x $0 ) ? () : ($^X) ), $0, @ARGV );
+}
+
+
 sub _process_request {
     my $self = shift;
 
@@ -284,12 +352,15 @@ sub _process_request {
         binmode STDIN,  ':raw';
         binmode STDOUT, ':raw';
 
-        my $remote_sockaddr = getpeername( $self->stdio_handle );
-        my ( undef, $iaddr ) = sockaddr_in($remote_sockaddr);
-        my $peeraddr = inet_ntoa($iaddr) || "127.0.0.1";
-
-        my ( $method, $request_uri, $proto ) = $self->parse_request;
+        # The ternary operator below is to protect against a crash caused by IE
+        # Ported from Catalyst::Engine::HTTP (Originally by Jasper Krogh and Peter Edwards)
+        # ( http://dev.catalyst.perl.org/changeset/5195, 5221 )
         
+        my $remote_sockaddr = getpeername( $self->stdio_handle );
+        my ( undef, $iaddr ) = $remote_sockaddr ? sockaddr_in($remote_sockaddr) : (undef,undef);
+        my $peeraddr = $iaddr ? ( inet_ntoa($iaddr) || "127.0.0.1" ) : '127.0.0.1';
+        
+        my ( $method, $request_uri, $proto ) = $self->parse_request;
         
         unless ($self->valid_http_method($method) ) {
             $self->bad_request;
@@ -299,12 +370,12 @@ sub _process_request {
         $proto ||= "HTTP/0.9";
 
         my ( $file, $query_string )
-            = ( $request_uri =~ /([^?]*)(?:\?(.*))?/ );    # split at ?
+            = ( $request_uri =~ /([^?]*)(?:\?(.*))?/s );    # split at ?
 
         $self->setup(
             method       => $method,
             protocol     => $proto,
-            query_string => ( $query_string || '' ),
+            query_string => ( defined($query_string) ? $query_string : '' ),
             request_uri  => $request_uri,
             path         => $file,
             localname    => $self->host,
@@ -326,13 +397,9 @@ sub _process_request {
         $self->post_setup_hook if $self->can("post_setup_hook");
 
         $self->handler;
-        }
+    }
 }
 
-
-
-
-
 =head2 stdio_handle [FILEHANDLE]
 
 When called with an argument, sets the socket to the server to that arg.
@@ -352,9 +419,9 @@ sub stdio_handle {
 
 =head2 stdin_handle
 
-Returns a filehandle used for input from the client.  By default, 
-returns whatever was set with C<stdio_handle>, but a subclass
-could do something interesting here (see L<HTTP::Server::Simple::Logger>).
+Returns a filehandle used for input from the client.  By default,
+returns whatever was set with C<stdio_handle>, but a subclass could do
+something interesting here.
 
 =cut
 
@@ -367,7 +434,7 @@ sub stdin_handle {
 
 Returns a filehandle used for output to the client.  By default, 
 returns whatever was set with C<stdio_handle>, but a subclass
-could do something interesting here (see L<HTTP::Server::Simple::Logger>).
+could do something interesting here.
 
 =cut
 
@@ -395,7 +462,7 @@ sub handler {
         croak "do not call " . ref($self) . "::SUPER->handler";
     }
     else {
-        die "handler called out of context";
+        croak "handler called out of context";
     }
 }
 
@@ -460,17 +527,34 @@ sub headers {
     my $headers = shift;
 
     my $can_header = $self->can("header");
+    return unless $can_header;
     while ( my ( $header, $value ) = splice @$headers, 0, 2 ) {
-        if ($can_header) {
-            $self->header( $header => $value );
-        }
+        $self->header( $header => $value );
     }
 }
 
 =head2 accept_hook
 
 If defined by a sub-class, this method is called directly after an
-accept happens.
+accept happens.  An accept_hook to add SSL support might look like this:
+
+    sub accept_hook {
+        my $self = shift;
+        my $fh   = $self->stdio_handle;
+
+        $self->SUPER::accept_hook(@_);
+
+        my $newfh =
+        IO::Socket::SSL->start_SSL( $fh, 
+            SSL_server    => 1,
+            SSL_use_cert  => 1,
+            SSL_cert_file => 'myserver.crt',
+            SSL_key_file  => 'myserver.key',
+        )
+        or warn "problem setting up SSL socket: " . IO::Socket::SSL::errstr();
+
+        $self->stdio_handle($newfh) if $newfh;
+    }
 
 =head2 post_setup_hook
 
@@ -490,7 +574,7 @@ yourself in subclasses.
 sub print_banner {
     my $self = shift;
 
-    print(    __PACKAGE__
+    print( ref($self) 
             . ": You can connect to your server at "
             . "http://localhost:"
             . $self->port
@@ -500,11 +584,8 @@ sub print_banner {
 
 =head2 parse_request
 
-Parse the HTTP request line.
-
-Returns three values, the request method, request URI and the protocol
-Sub-classed versions of this should return three values - request
-method, request URI and proto
+Parse the HTTP request line.  Returns three values, the request
+method, request URI and the protocol.
 
 =cut
 
@@ -528,13 +609,9 @@ sub parse_request {
 
 =head2 parse_headers
 
-Parse incoming HTTP headers from STDIN.
-
-Remember, this is a B<simple> HTTP server, so nothing intelligent is
-done with them C<:-)>.
-
-This should return an ARRAY ref of C<(header =E<gt> value)> pairs
-inside the array.
+Parses incoming HTTP headers from STDIN, and returns an arrayref of
+C<(header =E<gt> value)> pairs.  See L</headers> for possibilities on
+how to inspect headers.
 
 =cut
 
@@ -547,7 +624,7 @@ sub parse_headers {
     while ( sysread( STDIN, my $buff, 1 ) ) {
         if ( $buff eq "\n" ) {
             $chunk =~ s/[\r\l\n\s]+$//;
-            if ( $chunk =~ /^([\w\-]+): (.+)/i ) {
+            if ( $chunk =~ /^([^()<>\@,;:\\"\/\[\]?={} \t]+):\s*(.*)/i ) {
                 push @headers, $1 => $2;
             }
             last if ( $chunk =~ /^$/ );
@@ -569,8 +646,7 @@ sub setup_listener {
     my $self = shift;
 
     my $tcp = getprotobyname('tcp');
-
-    socket( HTTPDaemon, PF_INET, SOCK_STREAM, $tcp ) or die "socket: $!";
+    socket( HTTPDaemon, PF_INET, SOCK_STREAM, $tcp ) or croak "socket: $!";
     setsockopt( HTTPDaemon, SOL_SOCKET, SO_REUSEADDR, pack( "l", 1 ) )
         or warn "setsockopt: $!";
     bind( HTTPDaemon,
@@ -582,15 +658,16 @@ sub setup_listener {
             )
         )
         )
-        or die "bind: $!";
-    listen( HTTPDaemon, SOMAXCONN ) or die "listen: $!";
+        or croak "bind to @{[$self->host||'*']}:@{[$self->port]}: $!";
+    listen( HTTPDaemon, SOMAXCONN ) or croak "listen: $!";
 
 }
 
 
 =head2 after_setup_listener
 
-This method is called immediately after setup_listener. It's here just for you to override.
+This method is called immediately after setup_listener. It's here just
+for you to override.
 
 =cut
 
@@ -604,7 +681,7 @@ request was invalid.
 
 =cut
 
-our $bad_request_doc = join "", <DATA>;
+$bad_request_doc = join "", <DATA>;
 
 sub bad_request {
     my $self = shift;
@@ -617,19 +694,21 @@ sub bad_request {
 =head2 valid_http_method($method)
 
 Given a candidate HTTP method in $method, determine if it is valid.
-Override if, for example, you'd like to do some WebDAV.
+Override if, for example, you'd like to do some WebDAV.  The default
+implementation only accepts C<GET>, C<POST>, C<HEAD>, C<PUT>, and
+C<DELETE>.
 
 =cut 
 
 sub valid_http_method {
     my $self   = shift;
-    my $method = shift;
+    my $method = shift or return 0;
     return $method =~ /^(?:GET|POST|HEAD|PUT|DELETE)$/;
 }
 
 =head1 AUTHOR
 
-Copyright (c) 2004-2006 Jesse Vincent, <jesse@bestpractical.com>.
+Copyright (c) 2004-2008 Jesse Vincent, <jesse@bestpractical.com>.
 All rights reserved.
 
 Marcus Ramberg <drave@thefeed.no> contributed tests, cleanup, etc
@@ -637,6 +716,8 @@ Marcus Ramberg <drave@thefeed.no> contributed tests, cleanup, etc
 Sam Vilain, <samv@cpan.org> contributed the CGI.pm split-out and
 header/setup API.
 
+Example section by almut on perlmonks, suggested by Mark Fuller.
+
 =head1 BUGS
 
 There certainly are some. Please report them via rt.cpan.org
